from .constants import FETCH_GROUPS, GROUP_DEVICE, GROUP_DEVICE_REQ_INPUT, PASSED, FAILED
from .utilities import generate_error_list, validate_input
from .logger import logger

import os
import time
import math
import json
import asyncio

def fetch_all_groups(central_api_req):
    """
    Get a list of available groups from Central
        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
        Returns:
                group_list (list) : list of available group 
    """
    logger.debug("In function to fetch all groups")
    group_list = []
    try:
        api_path = FETCH_GROUPS
        method = "GET"
        limit = 20
        prev_offset = 0
        site_limits = {"offset": prev_offset, "limit": limit}
        while True:
            resp = central_api_req.request(method, api_path, params=site_limits)
            if resp.status_code == 200 or resp.status_code == 201:
                resp_data = resp.json()
                logger.info("Group data is available. Collecting the group data.")
                for site_details in resp_data["data"]:
                    group_list.extend(site_details)
                if resp_data["total"] > (prev_offset+limit):
                    prev_offset += limit
                    site_limits = {"offset": prev_offset, "limit": limit}
                    continue
                else:
                    break
            else:
                error_msg = f"HTTP Failure!. Status code : {resp.status_code}"
                logger.error(error_msg) 
        return group_list

    except Exception as e:
        logger.debug("Error while fetching all the groups from central")

        
def group_devices(central_api_req, input_data, **kwargs):
    """
    Associate devices to groups
        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list(dict)): list of device serial and device variables to be uploaded
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list(dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        group_list = fetch_all_groups(central_api_req)
        if len(input_data) > 0:
            if validate_input(GROUP_DEVICE_REQ_INPUT, input_data) != []:
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            group_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for data in input_data[index*limit:(index+1)*limit]:
                    if data.get("group") in group_list:
                        subtasks.append(
                            moving_devices_to_group(central_api_req, data.get("serial"), data.get("group")))
                    else:
                        group_name = data.get("group_name")
                        error_msg = f"Group {group_name} not created on central"
                        group_device_output_list.extend([{"serial": data.get("serial"), "status": FAILED, "message": error_msg}])
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                group_device_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return group_device_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)
    except Exception as e:
        error_msg = f"Error in associating devices to groups"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)


def group_device_to_default_group(central_api_req, input_data, **kwargs):
    """
    Associate devices to groups
        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list(dict)): list of device serial and device variables to be uploaded
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list(dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        group_list = fetch_all_groups(central_api_req)
        if len(input_data) > 0:
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            group_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for data in input_data[index*limit:(index+1)*limit]:
                    if data.get("group") in group_list:
                        subtasks.append(
                            moving_devices_to_group(central_api_req, data.get("serial"), "default"))
                    else:
                        group_name = data.get("group_name")
                        error_msg = f"Group {group_name} not created on central"
                        group_device_output_list.extend([{"serial": data.get("serial"), "status": FAILED, "message": error_msg}])
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                group_device_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return group_device_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)
    except Exception as e:
        error_msg = f"Error in unassociating devices to groups"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)


async def moving_devices_to_group(central_api_req, serial, group):
    """
	Asynchronous function to assign group to a device

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
				serial (str) : serial number of the selected device
				group (str) : group name to be associated

        Returns:
                dict : includes serial(str), status(Passed or Failed), ouput(str)
	"""
    try:
        api_path = GROUP_DEVICE
        method = "POST"
        group_data = {"group": group, "serials": [serial]}
        resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps(group_data))
        if (resp.status == 200) or (resp.status == 201):
            success_msg = f"{resp.status} - {serial} has successfully moved to group {group}"
            logger.info(success_msg)
            return {"serial": serial, "status": FAILED, "message": success_msg}
        else:
            error_msg = f"{resp.status} -  Failed to associated to group {group}"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        error_msg = f"{resp.status} -  Failed to associated to group {group}"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}

async def synctask(subtasks):
    task_level_output = await asyncio.gather(*subtasks)
    return task_level_output


