from .constants import UPLOAD_VARIABLE_API, IAP_GROUP_CHECK, SWITCH_GROUP_CHECK, UPLOAD_VARIABLES_REQ_INPUT, REMOVE_VARIABLES_REQ_INPUT, PASSED, FAILED
from .utilities import generate_error_list, validate_input
from .logger import logger

import os
import time
import math
import json
import asyncio

def iap_group_verify(central_api_req, serial):
    """
	Check the device is part of a central group

        Parameters:
			central_api_req (CentralAPIRequest) : Request object to make HTTP call
			serial (str) : serial number of the device

        Returns:
			Passed or Failed depending on the devices associated to group or not

	"""
    try:
        method = "GET"
        api_path = IAP_GROUP_CHECK.format(serial)
        resp = central_api_req.request(method, api_path)
        if resp.status_code == 200:
            resp_content = json.loads(resp.content)
            group_name = resp_content.get("group_name")
            if group_name:
                return PASSED
            else:
                return FAILED
        else:
            logger.info(f"{serial} - is not part of any group")
            return FAILED
    except Exception as e:
        logger.error(f"{serial} - error while checking the device group")


def switch_group_verify(central_api_req, serial):
    """
	Check the device is part of a central group

        Parameters:
			central_api_req (CentralAPIRequest) : Request object to make HTTP call
			serial (str) : serial number of the device

        Returns:
			Passed or Failed depending on the devices associated to group or not

	"""
    try:
        method = "GET"
        api_path = SWITCH_GROUP_CHECK.format(serial)
        resp = central_api_req.request(method, api_path)
        if resp.status_code == 200:
            resp_content = json.loads(resp.content)
            group_name = resp_content.get("group_name")
            if group_name:
                return PASSED
            else:
                return FAILED
        else:
            logger.info(f"{serial} - is not part of any group")
            return FAILED
    except Exception as e:
        logger.error(f"{serial} - error while checking the device group")

        
def upload_variables(central_api_req, input_data, **kwargs):
    """
    Upload variable list to devices

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list(dict)): list of device serial and device variables to be uploaded
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list (dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        if len(input_data) > 0:
            if validate_input(REMOVE_VARIABLES_REQ_INPUT, input_data) != []:
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            upload_variable_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for data in input_data[index*limit:(index+1)*limit].to_dict(orient='index').values():
                    if data.get("variables"):
                        subtasks.append(upload_variables_to_group(central_api_req, data.get("serial"), data.get("variables")))
                    else:
                        error_msg = f"Variables not provided for Uploading variables"
                        upload_variable_output_list.extend([{"serial": data.get("serial"), "status": FAILED, "message": error_msg}])
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                upload_variable_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return upload_variable_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)
    except Exception as e:
        error_msg = f"Error in uploading variables"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)


async def upload_variables_to_group(central_api_req, serial, variables):
    """
	Asynchronous function to upload device variables

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
				serial (str) : serial number of the device
                variables (dict) : Variables need to be uploaded

        Returns:
                dict: includes serial(str), status(Passed or Failed), ouput(str)

	"""
    try:
        api_path = UPLOAD_VARIABLE_API.format(serial)
        method = "POST"
        resp, resp_data = await central_api_req.asyncrequest("GET", api_path)
        if resp.status == 200 and resp_data != {}:
            resp, resp_data = await central_api_req.asyncrequest("DELETE", api_path)
            if resp.status == 200:
                await upload_variables(central_api_req, serial, variables)
        else:
            await upload_variables(central_api_req, serial, variables)
        
    except Exception as e:
        logger.error(e)
        error_msg = f"error occurred while licensing the device"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}

async def upload_variables(central_api_req, serial, data):
    try:
        api_path = UPLOAD_VARIABLE_API.format(serial)
        resp, resp_data = await central_api_req.asyncrequest("POST", api_path, data=json.dumps(data))
        if (resp.status == 200) or (resp.status == 201):
            success_message = f"{resp.status} - Success!! variables uploaded to {serial}"
            logger.info(success_message)
            return {"serial": serial, "status": PASSED, "message": success_message}
        else:
            error_msg = f"{resp.status} - Error while uploading variables"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        error_msg = "Error occurred while uploading the variables"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}

def remove_variables(central_api_req, input_data, **kwargs):
    """
    Upload variable list to devices

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list(dict)): list of device serial for which variables need to be removed
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list (dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        if len(input_data) > 0:
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            upload_variable_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for data in input_data[index*limit:(index+1)*limit].to_dict(orient='index').values():
                    subtasks.append(delete_variables(central_api_req, data.get("serial")))
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                upload_variable_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return upload_variable_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)
    except Exception as e:
        error_msg = f"Error in uploading variables"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)


async def delete_variables(central_api_req, serial):
    """
	Asynchronous function to delete variables uploaded to a device

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
				serial : serial number of the device

        Returns:
                dict: includes serial(str), status(Passed or Failed), ouput(str)

	"""
    try:
        api_path = UPLOAD_VARIABLE_API.format(serial_number)
        resp, resp_data = await central_api_req.asyncrequest("DELETE", api_path)
        if resp.status == 200:
            success_message = "ROLLBACK!! variables removed"
            logger.info(success_message)
            return {"serial": serial, "status": PASSED, "message": success_message}
        else:
            error_msg = "Failed to remove variables"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        error_msg = "Failed to remove variables"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}


async def synctask(subtasks):
    task_level_output = await asyncio.gather(*subtasks)
    return task_level_output

