from .constants import ADD_DEVICES, INVENTORY_VERIFY, ADD_DEVICES_REQ_INPUT, PASSED, FAILED
from .utilities import generate_error_list, validate_input
from .logger import logger

import os
import time
import math
import json
import asyncio


def add_device_using_mac_serial(central_api_req, input_data, **kwargs):
    """
    Adding device to the Central Account

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            input_data (list(dict)): Input data(mac and serial) to add devices
            **kwargs: Additional or Optional arguments
                        limit (default 20): limit for concurrent executions for asyncio
                        time_sleep (default 20s): time sleep for each async execution

        Returns:
            list(dict): includes serial(str), status(Passed or Failed), ouput(str)
    """
    try:
        input_list = []
        for each_input_dict in input_data:
            if ('mac' and 'serial') in each_input_dict.keys():
                input_list.append(each_input_dict)
            else:
                raise Exception("Invalid request data")
        input_data = input_list
        limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
        add_device_output_list = []
        if len(input_data) > 0:
            if validate_input(ADD_DEVICES_REQ_INPUT, input_data):
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            for index in range(math.ceil(len(input_data) / limit)):
                subtasks = []
                for device_info in input_data[index * limit:(index + 1) * limit]:
                    subtasks.append(add_device(central_api_req, device_info))
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                add_device_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index * limit:(index + 1) * limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return add_device_output_list
        else:
            error_message = "0 device selected to perform the task"
            logger.error(error_message)
            error_list = generate_error_list(input_data, FAILED, error_message)
            return error_list
    except Exception as e:
        logger.error(e)
        error_message = "Error adding the device to central inventory"
        logger.error(error_message)
        error_list = generate_error_list(input_data, FAILED, error_message)
        return error_list


async def add_device(central_api_req, device_info):
    """
    Asynchronous function to add device to central

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                device_info (dict): Input data

        Returns:
            dict: includes serial(str), status(Passed or Failed) and ouput(str)
    """
    try:
        serial = device_info.get('serial')
        method = "POST"
        api_path = ADD_DEVICES
        resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps([device_info]))
        if resp.status == 200:
            success_message = f"{serial} - Device added"
            logger.info(f"{resp.status} - Device added")
            return {"serial": serial, "status": PASSED, "message": success_message}
        elif resp.status == 429:
            error_message = f"{resp.status} Central API Error - API Rate limit exceeded"
            logger.error(error_message)
            return {"serial": serial, "status": FAILED, "message": error_message}
            
        elif resp.status == 500:
            error_message = f"{resp.status} Central API Error - Internal server error"
            logger.error(error_message)
            return {"serial": serial, "status": FAILED, "message": error_message}
            
        elif resp.status == 400 and resp_data.get("code") == "ATHENA_ERROR_NO_DEVICE" and resp_data.get("extra").get("message").get("invalid_device") != []:
            if resp_data.get("extra").get("message").get("invalid_device")[0].get("status") == "INVALID_MAC_SN":
                error_message = f"{resp.status} - INVALID_MAC_SN Invalid MAC or Serial "
                logger.error(error_message)
                return {"serial": serial, "status": FAILED, "message": error_message}
            else:
                error_message = f"{resp.status} - Failed to add devices to central"
                logger.error(error_message)
                return {"serial": serial, "status": FAILED, "message": error_message}
        
        elif resp.status == 400 and resp_data.get("code") == "ATHENA_ERROR_NO_DEVICE" and resp_data.get("extra").get("message").get("blocked_devices") != []:
            if resp_data.get("extra").get("message").get("blocked_device")[0].get("status") == "ATHENA_ERROR_DEVICE_ALREADY_EXIST":
                success_msg = f"{resp.status} - Device already Exists"
                logger.error(success_msg)
                return {"serial": serial, "status": PASSED, "message": success_msg}
                
        else:
            error_message = f"Failed to add device into central"
            logger.error(error_message)
            return {"serial": serial, "status": FAILED, "message": error_message}
            
    except Exception as e:
        logger.error(e)
        error_message = f"Error in adding AP to inventory"
        logger.error(error_message)
        return {"serial": serial, "status": FAILED, "message": error_message}
        
async def synctask(subtasks):
    task_level_output = await asyncio.gather(*subtasks)
    return task_level_output


def verify_device_addition(central_api_req, input_data, **kwargs):
    """
	Check the device whether it has joined central

        Parameters:
			central_api_req (CentralAPIRequest) : Request object to make HTTP call
			input_data (list(dict)) : Input data(mac and serial) to verify if devices added

        Returns:
			list(dict): includes serial(str), status(Passed or Failed), ouput(str)

	"""
    try:
        input_list = []
        for each_input_dict in input_data:
            if ('serial') in each_input_dict.keys():
                input_list.append(each_input_dict)
            else:
                raise Exception("Invalid request data")
        method = "POST"
        api_path = INVENTORY_VERIFY
        verify_device_output_list = []
        for data in input_list:
            resp = central_api_req.request(method, api_path, data=json.dumps([data]))
            resp_data = resp.json()
            device_serial = data.get("serial")
            if resp.status_code == 200 and len(resp_data.get("available_devices")) != 0:
                success_msg = f"{resp.status_code} - {device_serial} - is in the inventory..."
                logger.info(success_msg)
                verify_device_output_list.append({"serial": device_serial, "status": PASSED, "message": success_msg})
            else:
                error_msg = f"{resp.status_code} - {device_serial} - is not added to central inventory"
                logger.error(error_msg)
                verify_device_output_list.append({"serial": device_serial, "status": FAILED, "message": error_msg})
        return verify_device_output_list
    except Exception as e:
        error_msg = f"Error while verifying the devices addition"
        logger.error(error_msg)
        error_list = generate_error_list(input_data, FAILED, error_msg)
        return error_list

