from .constants import ADD_LICENSE, DELETE_LICENSE, CHECK_AUTOLICENSING, LICENSE_DEVICES_REQ_INPUT, PASSED, FAILED
from .utilities import generate_error_list, validate_input
from .logger import logger

import os
import time
import math
import json
import asyncio


def check_auto_license(central_api_req, **kwargs):
    """
    Check AutoLicensing

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            **kwargs: Additional or Optional arguments

        Returns:
            None: if autolicensed or services: services which are auto-licensed
    """
    try:
        api_path = CHECK_AUTOLICENSING
        method = "GET"
        auto_license_data = central_api_req.request(method, api_path)
        if auto_license_data.status_code == 200:
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            auto_services = auto_license_data.json().get("services")
            return auto_services
        else:
            logger.error("Error while checking the autolicensed services")
            return None
    except Exception as e:
        logger.error(e)
        error_message = "Error while checking the autolicensed services"
        logger.error(error_message)
        return None


def license_devices(central_api_req, input_data, **kwargs):
    """
    Assigning license to devices

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            input_data (json): Input data(mac and serial) to add devices
            **kwargs: Additional or Optional arguments
                        limit (default 20): limit for concurrent executions for asyncio
                        time_sleep (default 20s): time sleep for each async execution

        Returns:
            list(dict): includes serial(str), status(Passed or Failed), ouput(str)
    """
    try:
        auto_services = check_auto_license(central_api_req)
        if auto_services != None and len(input_data) > 0:
            if validate_input(LICENSE_DEVICES_REQ_INPUT, input_data):
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            license_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for license_details in input_data[index*limit:(index+1)*limit]:
                    auto_license_disabled_srvs = set(license_details.get("services", [])) - set(auto_services)
                    if auto_license_disabled_srvs == {}:
                        _services_ =  license_details.get("services", [])
                        success_message = f"Auto-licensing enabled for the services - {_services_}"
                        logger.info(success_message)
                        license_device_output_list = generate_error_list(input_data, PASSED, success_message)
                        return license_device_output_list
                    else:
                        auto_license_srvs = set(license_details.get("services", [])) - auto_license_disabled_srvs
                        logger.info(f"Auto-Licensing enabled for the services - {auto_license_srvs}")
                        if auto_license_disabled_srvs != {}:
                            subtasks.append(assign_subscription_device(
                                central_api_req, license_details.get("serial"), list(auto_license_disabled_srvs)))
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                license_device_output_list.extend(asyncio.run(synctask(subtasks)))

                if len(input_data[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return license_device_output_list
        else:
            error_message = "Unable to check auto-license status or No devices selected to perform this task"
            logger.error(error_message)
            error_list = generate_error_list(input_data, FAILED, error_message)
            return error_list

    except Exception as e:
        error_message = "Error in licensing devices"
        logger.error(error_message)
        error_list = generate_error_list(input_data, FAILED, error_message)
        return error_list


async def assign_subscription_device(central_api_req, serial, services):
    """
	Asynchronous function to assign license to a device

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
				serial (str) : serial number of the selected row
                services (list): services need to be subscribed

        Returns:
                (dict): includes serial(str), status(Passed or Failed), ouput(str)

	"""
    try:
        # TODO: verify device addition on central
        api_path = ADD_LICENSE
        method = "POST"
        subscription_details = {
            "serials": [serial], "services": services
            }
        logger.info(f"device ready to subscribe - {subscription_details}")
        resp, resp_data = await central_api_req.asyncrequest(
            method, api_path, data=json.dumps(subscription_details)
        )
        if (resp.status == 200) or (resp.status == 201):
            success_msg = f"{resp.status} - Subscription tokens added successfully"
            logger.info(success_msg)
            return {"serial": serial, "status": PASSED, "message": success_msg}

        elif resp.status == 400 and resp_data.get("message") == "Invalid device":
            error_msg = f"{resp.status} - Invalid device"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}
            
        elif resp.status == 500:
            error_msg = f"{resp.status} - Internal Server Error, Server is not responding"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

        else:
            error_msg = f"{resp.status} - Failed License Task - Reason: API Response Failed "
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        logger.error(e)
        error_msg = f"error occurred while licensing the device"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}


# Rollback
def unlicense_devices(central_api_req, input_data, **kwargs):
    """
    Unassigning license to devices

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            input_data (json): Input data(mac and serial) to add devices
            **kwargs: Additional or Optional arguments
                        limit (default 20): limit for concurrent executions for asyncio
                        time_sleep (default 20s): time sleep for each async execution

        Returns:
            list(dict): includes serial(str), status(Passed or Failed), ouput(str)
    """
    try:
        if len(input_data) > 0:
            if validate_input(LICENSE_DEVICES_REQ_INPUT, input_data):
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            license_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for license_details in input_data[index*limit:(index+1)*limit]:
                    subtasks.append(unassign_subscription_device(
                        central_api_req, license_details.get("serial"), license_details.get("services", [])))
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                license_device_output_list.extend(asyncio.run(synctask(subtasks)))

                if len(input_data[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return license_device_output_list
        else:
            error_message = "0 devices selected to perform this task"
            logger.error(error_message)
            error_list = generate_error_list(input_data, FAILED, error_message)
            return error_list
    except Exception as e:
        error_message = "Error in unlicensing devices"
        logger.error(error_message)
        error_list = generate_error_list(input_data, FAILED, error_message)
        return error_list


async def unassign_subscription_device(central_api_req, serial, services):
    """
	Asynchronous function to unassign license from a device

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
				serial : serial number of the device
				license : license type to unassign to the device

        Returns:
                (dict): includes serial(str), status(Passed or Failed), ouput(str)

	"""
    try:
        api_path = DELETE_LICENSE
        method = "POST"
        subscription_details = {
            "serials": [serial], "services": services}
        resp, resp_data = await central_api_req.asyncrequest(
            method, api_path, data=json.dumps(subscription_details)
        )

        if (resp.status == 200) or (resp.status == 201):
            success_msg = f"Subscription tokens successfully deleted for {serial}"
            logger.info(success_msg)
            return {"serial": serial, "status": PASSED, "message": success_msg}
            
        else:
            error_msg = f"Unable to remove licensing to {serial}"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}
    except Exception as e:
        error_msg = f"Unable to remove licensing to {serial}"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}


async def synctask(subtasks):
    task_level_output = await asyncio.gather(*subtasks)
    return task_level_output

