from .constants import FETCH_AP_DETAILS, FETCH_SWITCH_DETAILS, SYNC_WAIT_SECONDS, PASSED, FAILED
from .utilities import generate_error_list
from .logger import logger

import os
import time
import math
import json
import asyncio

async def device_reachability_check(central_api_req, serial : str, **kwargs) -> dict:
    """
	Check if the device is online on central

        Parameters:
			central_api_req (CentralAPIRequest) : Request object to make HTTP call
			input_data (dict) : Input data(mac and serial) to verify if devices added
            **kwargs : additional arguments 
                        sync_time (default 60s) : wait time to check the device reachability 
                        wait_until (default 1500s): max threshold to wait for the device to come online

        Returns:
			(dict): includes serial(str), status(Passed or Failed), ouput(str)

	"""
    try:
        sync_time = kwargs.get('sync_time') if 'sync_time' in kwargs.keys() else 60
        wait_until = kwargs.get('wait_until') if 'wait_until' in kwargs.keys() else 1500
        while sync_time < wait_until:
            method = "GET"
            api_path = FETCH_AP_DETAILS.format(serial)
            resp = central_api_req.request(method, api_path)
            if resp.status_code == 200:
                resp_content = json.loads(resp.content)
                status = resp_content.get("status")
                if status == "Up":
                    success_msg = f"{resp.status_code} - Device is UP on central"
                    return {"serial": serial, "status": PASSED, "message": success_msg}
                else:
                    logger.info(f"{serial} - status is down, please wait while the device come up online on central.")
                    logger.info(f"Script will check the device status again in {SYNC_WAIT_SECONDS} seconds")
                    await asyncio.sleep(SYNC_WAIT_SECONDS)
                    sync_time += SYNC_WAIT_SECONDS
                    continue
            else:
                logger.info(f"{serial} - is not reachable, please wait while the device come up online.")
                logger.info(f"Script will check the device status again in {SYNC_WAIT_SECONDS} seconds")
                await asyncio.sleep(SYNC_WAIT_SECONDS)
                sync_time += SYNC_WAIT_SECONDS
                continue
        error_msg = f"Waited for {sync_time} seconds, still the device was not reachable"
        return {"serial": serial, "status": FAILED, "message": error_msg}
    except Exception as e:
        error_msg = f"{serial} - error while checking device reachability"
        return {"serial": serial, "status": FAILED, "message": error_msg}