from .constants import FETCH_SITES, SITE_DEVICE_API, FETCH_AP_DETAILS, SITE_DEVICE_REQ_INPUT, PASSED, FAILED
from .utilities import generate_error_list, validate_input
from .logger import logger

import os
import time
import math
import json
import asyncio

def fetch_all_sites(central_api_req):
    """
    Fetch all sites from Central

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call

        Returns:
            site_list (list) : List of all site details
    """
    try:
        api_path = FETCH_SITES
        method = "GET"
        site_list = []
        limit = 100
        prev_offset = 0
        site_limits = {"offset": prev_offset, "limit": limit}
        while True:
            resp = central_api_req.request(method, api_path, params=site_limits)
            if resp.status_code == 200:
                resp_data = resp.json()
                logger.info("Site data is available. Collecting the site data.")
                for site_details in resp_data.get("sites"):
                    site_list.append(site_details)

                if resp_data.get("total") > (prev_offset + limit):
                    prev_offset += limit
                    site_limits = {"offset": prev_offset, "limit": limit}
                    continue
                else:
                    break
            else:
                error_msg = f"HTTP Failure!. Status code : {resp.status_code}"
                logger.error(error_msg)
        return site_list

    except Exception as e:
        logger.debug("Error while fetching all the sites from central")


def get_site_by_serial(central_api_req, serial):
	try:
		method = "GET"
		api_path = FETCH_AP_DETAILS.format(serial)
		resp = central_api_req.request(method, api_path)
		if resp.status_code == 200:
			resp = resp.json()
			return resp.get("site_name")
		else:
			return None

	except Exception as e:
		logger.error(f"Error while fetching site details details by serial")

        
def add_iaps_to_site(central_api_req, input_data, **kwargs):
    """
    associate sites to devices

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list of dict): list of device serial and sites to be associated
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list(dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        sites_present = fetch_all_sites(central_api_req)
        available_sites = [site.get("site_name") for site in sites_present]
        if len(input_data) > 0:
            if validate_input(SITE_DEVICE_REQ_INPUT, input_data):
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            site_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for data in input_data[index*limit:(index+1)*limit]:
                    if data.get("site") in available_sites:
                        selected_site = \
                            [site for site in sites_present if site.get("site_name") == data.get("site")][0]
                        subtasks.append(associate_iap_site_device(central_api_req, data.get("serial"), selected_site.get("site_id")))
                    else:
                        error_msg = f"Error in associating devices to sites"
                        site_device_output_list.extend([{"serial": data.get("serial"), "status": FAILED, "message": error_msg}])
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                site_device_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return site_device_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)
    except Exception as e:
        error_msg = f"Error in associating devices to sites"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)


async def associate_iap_site_device(central_api_req, serial, site_id):
    """
    Asynchronous function to associate device to central

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            serial : serial number of the device
            site_id : central site ID of the site from input file

        Returns:
            (dict): includes serial(str), status(Passed or Failed), ouput(str)

    """
    try:
        api_path = SITE_DEVICE_API
        method = "POST"
        data = {
            "site_id": int(site_id),
            "device_type": "IAP",
            "device_ids": [serial]
        }
        resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps(data))
        if resp.status == 200:
            success_msg = f"{resp.status} - Successfully added IAP to site"
            logger.info(f"Successfully added IAP to site for {serial}")

            if resp_data.get("failed"):
                if resp_data.get("failed")[0].get("reason") == "SITE_ERR_MAX_NO_ALREADY_ASSIGNED":
                    resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps(data))
                    if resp.status == 200:
                        success_msg = f"{resp.status} - Successfully added IAP to site"
                        logger.info(success_msg)
                        return {"serial": serial, "status": PASSED, "message": success_msg}
                    else:
                        error_msg = f"{resp.status} - Failed to add switch to site"
                        logger.error(error_msg)
                        return {"serial": serial, "status": FAILED, "message": error_msg}

                elif resp_data.get("failed")[0].get("reason") == 'SITE_ERR_ALREADY_ASSOCIATED':
                    success_msg = f"{resp.status} - Site already associated"
                    return {"serial": serial, "status": PASSED, "message": success_msg}

                elif resp_data.get("failed")[0].get("reason") == 'SITE_ASSOC_ERR_DEVICE_NOT_PRESENT':
                    error_msg = f"{resp.status} - Site association error - Device not present"
                    return {"serial": serial, "status": FAILED, "message": error_msg}
                
                elif resp_data.get("failed")[0].get("reason") == 'SITE_ERR_DEFAULT':
                    error_msg = f"{resp.status} - Site Association Error, Central response - SITE_ERR_DEFAULT"
                    return {"serial": serial, "status": FAILED, "message": error_msg}

            return {"serial": serial, "status": PASSED, "message": success_msg}
        else:
            error_msg = f"{resp.status} - Failed to add IAP to site"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        logger.error(e)
        error_msg = f"Error while adding device to site"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}


def add_switches_to_site(central_api_req, input_data, **kwargs):
    """
    associate sites to switches

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list of dict): list of device serial and sites to be associated
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list(dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        sites_present = fetch_all_sites(central_api_req)
        available_sites = [site.get("site_name") for site in sites_present]
        if len(input_data) > 0:
            if validate_input(SITE_DEVICE_REQ_INPUT, input_data):
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            site_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for data in input_data[index*limit:(index+1)*limit]:
                    if data.get("site") in available_sites:
                        selected_site = \
                            [site for site in sites_present if site.get("site_name") == data.get("site")][0]
                        subtasks.append(associate_switch_site_device(central_api_req, data.get("serial"), selected_site.get("site_id")))
                    else:
                        error_msg = f"Error in associating devices to sites"
                        logger.error(error_msg)
                        return generate_error_list(input_data, FAILED, error_msg)
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                site_device_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(pd_device_details[index*limit:(index+1)*limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return site_device_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)
    except Exception as e:
        error_msg = f"Error in associating devices to sites"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)


async def associate_switch_site_device(central_api_req, serial, site_id):
    """
    Asynchronous function to associate device to central

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            serial (str) : serial number of the device
            site_id (int) : central site ID of the site from input file

        Returns:
            (dict): includes serial(str), status(Passed or Failed), ouput(str)

    """
    try:
        api_path = SITE_DEVICE_API
        method = "POST"
        data = {
            "site_id": int(site_id),
            "device_type": "SWITCH",
            "device_ids": [serial]
        }
        resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps(data))
        if resp.status == 200:
            success_msg = f"{resp.status} - Successfully added IAP to site"
            logger.info(f"Successfully added IAP to site for {serial}")

            if resp_data.get("failed"):
                if resp_data.get("failed")[0].get("reason") == "SITE_ERR_MAX_NO_ALREADY_ASSIGNED":
                    resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps(data))
                    if resp.status == 200:
                        success_msg = f"{resp.status} - Successfully added IAP to site"
                        logger.info(success_msg)
                        return {"serial": serial, "status": PASSED, "message": success_msg}
                    else:
                        error_msg = f"{resp.status} - Failed to add switch to site"
                        logger.error(error_msg)
                        return {"serial": serial, "status": FAILED, "message": error_msg}

                elif resp_data.get("failed")[0].get("reason") == 'SITE_ERR_ALREADY_ASSOCIATED':
                    success_msg = f"{resp.status} - Site already associated"
                    return {"serial": serial, "status": PASSED, "message": success_msg}

                elif resp_data.get("failed")[0].get("reason") == 'SITE_ASSOC_ERR_DEVICE_NOT_PRESENT':
                    error_msg = f"{resp.status} - Site association error - Device not present"
                    return {"serial": serial, "status": FAILED, "message": error_msg}
                
                elif resp_data.get("failed")[0].get("reason") == 'SITE_ERR_DEFAULT':
                    error_msg = f"{resp.status} - Site Association Error, Central response - SITE_ERR_DEFAULT"
                    return {"serial": serial, "status": FAILED, "message": error_msg}

            return {"serial": serial, "status": PASSED, "message": success_msg}
        else:
            error_msg = f"{resp.status} - Failed to add IAP to site"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        logger.error(e)
        error_msg = f"Error while adding device to site"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}


def iap_remove_site_device(central_api_req, input_data, **kwargs):
    """
    unassociate sites from aps

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list of dict): list of device serial and sites to be unassociated
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list(dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        if len(input_data) > 0:
            if validate_input(SITE_DEVICE_REQ_INPUT, input_data):
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            site_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for device in input_data[index*limit:(index+1)*limit]:
                    subtasks.append(
                        unassociate_iap_site_device(central_api_req, device.get("serial")))
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                site_device_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index * limit:(index + 1) * limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return site_device_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)

    except Exception as e:
        error_msg = f"Error in unassociating Site"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)

async def unassociate_iap_site_device(central_api_req, serial):
    """
    Asynchronous function to unassociate ap from central

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            serial (str) : serial number of the device

        Returns:
            (dict): includes serial(str), status(Passed or Failed), ouput(str)

    """
    try:
        method = "DELETE"
        api_path = SITE_DEVICE_API
        site_name = get_site_by_serial(central_api_req, serial)
        available_sites = fetch_all_sites(central_api_req)
        site_details = [site for site in available_sites if site.get(
            "site_name") == site_name]
        if site_details:
            data = {"site_id": int(site_details[0].get("site_id")), "device_type": "IAP",
                        "device_ids": [serial]}
            resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps(data))
            if resp.status == 200:
                success_msg = f"ROLLBACK!! Unassociated site"
                logger.info(success_msg)
                return {"serial": serial, "status": PASSED, "message": success_msg}
            else:
                error_msg = f"Error while unassociating sites of devices"
                logger.error(error_msg)
                return {"serial": serial, "status": FAILED, "message": error_msg}
        else:
            error_msg = f"Site not available to associate to a device {serial}"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        error_msg = f"Error in unassociating the site for the IAP device - {serial}"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}


def switch_remove_site_device(central_api_req, input_data, **kwargs):
    """
    unassociate site from switches

        Parameters:
                central_api_req (CentralAPIRequest) : Request object to make HTTP call
                input_data (list of dict): list of device serial and sites to be unassociated
                **kwargs: Additional or Optional arguments
                            limit (default 20): limit for concurrent executions for asyncio
                            time_sleep (default 20s): time sleep for each async execution

        Returns:
               list(dict): includes serial(str), status(Passed or Failed), ouput(str)
 

    """
    try:
        if len(input_data) > 0:
            if validate_input(SITE_DEVICE_REQ_INPUT, input_data):
                error_msg = ("Validation Failed for input list provided")
                return [{"serial": 'NA', "status": FAILED, "message": error_msg}]
            limit = kwargs.get('limit') if 'limit' in kwargs.keys() else 20
            site_device_output_list = []
            for index in range(math.ceil(len(input_data)/limit)):
                subtasks = []
                for device in input_data[index*limit:(index+1)*limit]:
                    subtasks.append(
                        unassociate_switch_site_device(central_api_req, device.get("serial")))
                if os.name == "nt":
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                site_device_output_list.extend(asyncio.run(synctask(subtasks)))
                if len(input_data[index * limit:(index + 1) * limit]) > 20:
                    logger.debug("Time sleep started")
                    time_sleep_in_secs = kwargs.get('time_sleep') if 'time_sleep' in kwargs.keys() else 20
                    time.sleep(time_sleep_in_secs)
                    logger.debug("Time sleep ended")
            return site_device_output_list
        else:
            error_msg = "0 device selected to perform the task"
            logger.error(error_msg)
            return generate_error_list(input_data, FAILED, error_msg)

    except Exception as e:
        error_msg = f"Error in unassociating Site"
        logger.error(error_msg)
        return generate_error_list(input_data, FAILED, error_msg)

async def unassociate_switch_site_device(central_api_req, serial):
    """
    Asynchronous function to unassociate switch from central

        Parameters:
            central_api_req (CentralAPIRequest) : Request object to make HTTP call
            serial (str) : serial number of the device
            site_id (int) : central site ID of the site from input file

        Returns:
            (dict): includes serial(str), status(Passed or Failed), ouput(str)

    """
    try:
        method = "DELETE"
        api_path = SITE_DEVICE_API
        site_name = get_site_by_serial(central_api_req, serial)
        available_sites = fetch_all_sites(central_api_req)
        site_details = [site for site in available_sites if site.get(
            "site_name") == site_name]
        if site_details:
            data = {"site_id": int(site_details[0].get("site_id")), "device_type": "SWITCH",
                        "device_ids": [serial]}
            resp, resp_data = await central_api_req.asyncrequest(method, api_path, data=json.dumps(data))
            if resp.status == 200:
                success_msg = f"ROLLBACK!! Unassociated site"
                logger.info(success_msg)
                return {"serial": serial, "status": PASSED, "message": success_msg}
            else:
                error_msg = f"Error while unassociating sites of devices"
                logger.error(error_msg)
                return {"serial": serial, "status": FAILED, "message": error_msg}
        else:
            error_msg = f"Site not available to associate to a device {serial}"
            logger.error(error_msg)
            return {"serial": serial, "status": FAILED, "message": error_msg}

    except Exception as e:
        error_msg = f"Error in deleting the IAP device - {serial}"
        logger.error(error_msg)
        return {"serial": serial, "status": FAILED, "message": error_msg}


async def synctask(subtasks):
    task_level_output = await asyncio.gather(*subtasks)
    return task_level_output
        
